import socket
import struct

MAX_TIME = 3


def make_socket_udp(num):
    proto_u = socket.getprotobyname('udp')
    udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, proto_u)
    udp_socket.setsockopt(socket.SOL_IP, socket.IP_TTL, num)
    return udp_socket


def make_socket_icmp(num, base_port):
    icmp_socket = socket.socket(socket.AF_INET, socket.SOCK_RAW,
                                socket.getprotobyname("icmp"))
    icmp_socket.setsockopt(socket.IPPROTO_IP, socket.IP_TTL,
                           struct.pack('I', num))
    icmp_socket.settimeout(MAX_TIME)
    icmp_socket.bind(('', base_port + num))
    return icmp_socket



class TryData:

    def __init__(self, count_of_try: int = 3):
        self.count_of_success = 0
        self.all_try = []
        for num_of_try in range(count_of_try):
            self.all_try.append(0.0)
        self.addr = ''

    def add(self, num_of_try: int, value: float):
        self.all_try[num_of_try] = value
        self.count_of_success += 1

    def get_midle_sum(self):
        sum_all = 0
        for value in self.all_try:
            sum_all += value
        if not self.count_of_success:
            return 0.0
        return sum_all / self.count_of_success



import socket
import struct
import time
import select
import os
from core.try_data import TryData
from core.print_method import print_data
from core.socket_maker import make_socket_icmp

BASE_TRY_NUM = 3
DEFAULT_COUNT_BYTE = 2048
MAX_TIME = 3
ICMP_ANSWER = 0
ICMP_TRACE = 11
ICMP_NO_ROUTE = 3
ICMP_HEAD = 8


def get_route(addr: str, hop: int, base_port: int) -> list:
    res_data = []
    complete = False

    for ttl in range(1, hop+1):
        if complete:
            print('Complete')
            break
        print(f'{str(ttl) + ")":<6}', end='')
        result = TryData()
        cur = None

        for num_of_try in range(BASE_TRY_NUM):
            with make_socket_icmp(ttl, base_port) as icmp_socket:
                icmp_socket.sendto(get_pack(), (addr, base_port + ttl))
                send_time = time.time()

                try:
                    select_socket = select.select([icmp_socket], [], [],
                                                  MAX_TIME)
                    if not select_socket[0]:
                        break
                    rec_pack, cur = icmp_socket.recvfrom(DEFAULT_COUNT_BYTE)
                    cur = cur[0]
                    result.addr = cur
                    recv_time = time.time()
                except socket.timeout:
                    continue
                else:

                    icmp_type, *_ = struct.unpack("bbHHh", rec_pack[20:28])
                    if icmp_type == ICMP_TRACE or icmp_type == ICMP_NO_ROUTE:
                        result.add(num_of_try, recv_time - send_time)
                    elif icmp_type == ICMP_ANSWER:
                        col = struct.calcsize("d")
                        s_time = struct.unpack("d", rec_pack[28:28 + col])[0]
                        result.add(num_of_try, recv_time - s_time)
                        complete = True
                    else:
                        break

        if not result.count_of_success:
            print('* * *')
        if cur is not None:
            print_data(result)
            res_data.append(result)
        if cur == addr:
            complete = True
        if ttl == hop:
            print('End of max hop')
    return res_data


def get_pack():
    os_id = os.getpid() & 0xFFFF
    head = struct.pack("bbHHh", ICMP_HEAD, 0, 0, os_id, 1)
    body = struct.pack("d", time.time())
    control_sum = socket.htons(get_sum(head + body))
    head = struct.pack("bbHHh", ICMP_HEAD, 0, control_sum, os_id, 1)
    return head + body


# Вычисление контрольной суммы для пакета
def get_sum(data):
    sum_of_data = 0
    len_of_operation = (len(data) // 2) * 2
    place = 0

    while place < len_of_operation:
        temp = data[place + 1] * 256 + data[place]
        place += 2
        sum_of_data = sum_of_data + temp
        sum_of_data = sum_of_data & 0xffffffff

    if len_of_operation < len(data):
        sum_of_data += ord(data[-1])
        sum_of_data = sum_of_data & 0xffffffff

    sum_of_data = (sum_of_data >> 16) + (sum_of_data & 0xffff)
    sum_of_data = sum_of_data + (sum_of_data >> 16)
    control = ~sum_of_data & 0xffff
    final = control >> 8 | (control << 8 & 0xff00)

    return final



import time
import select
from core.try_data import TryData
from core.print_method import print_data
from core.socket_maker import *

BASE_TRY_NUM = 3
DEFAULT_COUNT_BYTE = 1024
MAX_TIME = 3


def get_route(addr: str, hop: int, base_port: int) -> list:
    res_data = []
    complete = False
    for ttl in range(1, hop+1):
        if complete:
            print('Complete')
            break
        cur = None
        result = TryData()
        print(f'{str(ttl)+")":<6}', end='')

        for num_of_try in range(BASE_TRY_NUM):
            with make_socket_udp(ttl) as udp_socket,\
                    make_socket_icmp(ttl, base_port) as icmp_socket:
                udp_socket.sendto(''.encode(), (addr, base_port + ttl))
                send_time = time.time()
                try:
                    in_select_socket = select.select([icmp_socket], [], [],
                                                     MAX_TIME)
                    if not in_select_socket[0]:
                        break
                    _, cur = icmp_socket.recvfrom(DEFAULT_COUNT_BYTE)
                    recv_time = time.time()
                    cur = cur[0]
                    result.addr = cur
                    result.add(num_of_try, recv_time-send_time)
                except socket.timeout:
                    continue

        if not result.count_of_success:
            print('* * *')

        if cur is not None:
            print_data(result)
            res_data.append(result)

        if cur == addr:
            print('Complete')
            break

        if ttl == hop:
            print("End of max hop")
            break
    return res_data



from core.try_data import TryData
import socket


def print_data(result: TryData):
    try:
        name = socket.gethostbyaddr(result.addr)[0]
    except socket.herror:
        name = 'Unknown server'
    finally:
        f_time = round(result.get_midle_sum() * 1000, 2)
        print(f'{result.addr:<15}    {name:<50}   {f_time} ms')



import socket
import matplotlib.pyplot as plt
import matplotlib
import core.udp_trace as udp_trace
import sys
import core.icmp_trace as icmp_trace
import argparse
from core.try_data import TryData

TRACE_METHOD = udp_trace.get_route
STANDART_DPI = 80


def main():

    parser = argparse.ArgumentParser(description='traceroute argument')
    parser.add_argument('addr', type=str, help='dest addr')
    parser.add_argument('hop', type=str, help='max hop count')
    parser.add_argument('-g', action='store_true', help='make graf')
    parser.add_argument('-I', '--ICMP', action='store_true',
                        help='use ICMP method')
    parser.add_argument('-p', '--base_port',
                        default='33434', help='give base port')
    namespaces = parser.parse_args()
    if sys.platform == 'win32':
        print('work on this platform is not supported')
        exit(0)
    if not check_correct(namespaces):
        print('Please use correct input')
        exit(0)
    addr = socket.gethostbyname(namespaces.addr)
    hop = int(namespaces.hop)
    res_data = TRACE_METHOD(addr, hop, int(namespaces.base_port))

    if namespaces.g:
        get_graf(res_data)


def check_correct(argv):
    try:

        global TRACE_METHOD
        if argv.ICMP:
            TRACE_METHOD = icmp_trace.get_route
        socket.gethostbyname(argv.addr)
        if int(argv.hop) > 255:
            return False
        int(argv.base_port)

    except socket.gaierror:
        print('Not correct dest adr')
        return False

    except ValueError:
        print('Not correct hop count')
        return False

    return True


def get_graf(res_data):
    if res_data == []:
        result = TryData()
        result.ddr = 'No_Data'
        res_data.append(result)
    dpi = STANDART_DPI
    coef = round(len(res_data) / 25)
    if coef == 0:
        coef = 1
    fig = plt.figure(dpi=dpi, figsize=(512 * coef / dpi, 480 * coef / dpi))
    matplotlib.rcParams.update({'font.size': 8})
    plt.title('Traceroute')

    xs = range(len(res_data))

    plt.bar([x + 0.05 for x in xs], [v.all_try[0] for v in res_data],
            width=0.2, color='red', alpha=0.7, label='first try',
            zorder=2)
    plt.bar([x + 0.3 for x in xs], [v.all_try[1] for v in res_data],
            width=0.2, color='blue', alpha=0.7, label='second try',
            zorder=2)
    plt.bar([x + 0.55 for x in xs], [v.all_try[2] for v in res_data],
            width=0.2, color='green', alpha=0.7, label='third try',
            zorder=2)
    plt.xticks(xs, [d.addr for d in res_data])

    fig.autofmt_xdate(rotation=90)

    plt.legend(loc='upper right')
    fig.savefig('trace.svg')


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

setup(name='li-traceroute',

      version='1.2',

      url='https://github.com/ShapeShifter420/li-traceroute',

      description='python simple traceroute ICMP/UDP+ICMP for linux',

      packages=['core'],

      long_description=open('README.md').read(),

      test_suite='test',

      install_requires=['matplotlib'],

      entry_points={
        'console_scripts': ['li-traceroute = core.traceroute:main']
        }
      )
